#!/usr/bin/perl -w

# Program dirparser.pl

# (c) Gary C. Kessler, 2011 [gck@garykessler.net]

# Program to parse FAT directory data structures.
# Ref: File System Forensic Analysis, Brian Carrier

# ********************************************************
# MAIN PROGRAM BLOCK
# ********************************************************

#use Encode qw/encode_utf8 decode_utf8/;
use utf8;
binmode STDOUT, ":utf8";

# Initialize global variables

$version = "2.1";
$build_date = "3 January 2018";

# $source is the name of the file to read.

$source = "";

# The attribute_values is the bit-flag mapping for the attribute
# byte in a short file name. An attribute byte value of 0x0F
# indicates a long file name.

%attribute_values =
  (
  "0x01"=>"Read-only",
  "0x02"=>"Hidden",
  "0x04"=>"System",
  "0x08"=>"Volume label",
  "0x10"=>"Directory",
  "0x20"=>"Archive"
  );

# lfn_pos is the position of the low-end byte of each 16-bit
# character entry in an LFN

@lfn_pos = (1, 3, 5, 7, 9, 14, 16, 18, 20, 22, 24, 28, 30);

#
# Begin program activity!!
#

print "\nDirectory Entry Parser V$version - Gary C. Kessler ($build_date)\n\n";

# Parse command for file name and request missing information.

if (parsecommandline ())
  { exit 1; };

print "Source file = $source\n";

# Open input ($source) file and stop program if error

open (INFILE, "<:raw", $source)
  or die "Can't open input file \"$source\": $!";
binmode INFILE;

# Start by reading the entire file. Set the max at 64 sectors (32,768 bytes)

$dir_len = 0;
while (not (eof INFILE) && $dir_len < 32768)
  {
  read (INFILE,$byte,1);
  $dir_file [$dir_len] = unpack ("C", $byte);
  $dir_len++;
  }

close INFILE;
printf ("Input file length = %d bytes (%d sector(s)) and contains %d entries.\n\n", $dir_len, $dir_len/512, $dir_len/32);

if ($dir_len < 512)
  { die "File is smaller than one sector ($dir_len bytes)\n"; }
if ($dir_len%512 != 0)
  { die "File is not an integer number of sectors ($dir_len bytes)\n"; }

# Loop through all entries. Each directory entry is 32 bytes in length.

$num_entries = $dir_len / 32;
for ($entry = 1; $entry <= $num_entries; $entry++)
  {

# bbase is the byte position in the file

  $bbase = ($entry - 1) * 32;

# Process based on LFN (attribute = 0x0F) or SFN; empty entry (1st byte = 0x00)
#  vs. non-empty entry.

# Case 1: First byte = 0x00 means we're done.

  if ($dir_file[$bbase] == 0x00)
      {
      printf("*** Entries %d-%d (bytes %d-%d) should be empty ***\n", $entry,$num_entries,
         $bbase, $dir_len-1);
      $nlines = ($num_entries - $entry + 1) * 2;
      for ($i=0; $i<$nlines; $i++)
        {
        set_output ($bbase,16,@dir_file);
        $bbase += 16;
        }
      last;
      }
    else
      {

# Case 2: Long file name entry. Note that LFNs might occupy more than one
#   directory entry, so might re-adjust $entry

      if ($dir_file [$bbase+11] == 0x0f)
          { $entry = process_LFN (); }

# Case 3: Short file name entry

        else
          {
          process_SFN ();
          }
      }
  }

# Finish up!

print "\nDone!\n";


# ********************************************************
# *****           SUBROUTINE BLOCKS                  *****
# ********************************************************

# ********************************************************
# commify
# ********************************************************

# Routine from PerlFAQ5 -- add commas to large integers
# http://perldoc.perl.org/perlfaq5.html
  #How-can-I-output-my-numbers-with-commas-added?

sub commify
{
local $_  = shift;
1 while s/^([-+]?\d+)(\d{3})/$1,$2/;
return $_;
}

# ********************************************************
# get_date
# ********************************************************

# Unpack DOS/FAT date stored in $date.
# Day = bits 0:4, Month = bits 5:8, Year = bits 9:15 + 1980

sub get_date
{
my ($date) = @_;
my ($mon, $day, $year, $temp);

$temp = $date;

$day = $temp % 32;
$temp = int ($temp / 32);
$mon = $temp % 16;
$year = int ($temp / 16) + 1980;

return ($mon, $day, $year);
}

# ********************************************************
# get_time
# ********************************************************

# Unpack DOS/FAT time stored in $time.
# Seconds = bits 0:4 * 2, Minute = bits 5:10, Hour = bits 11:15

sub get_time
{
my ($time) = @_;
my ($hour, $min, $sec, $temp);

$temp = $time;

$sec = $temp % 32 * 2;
$temp = int ($temp / 32);
$min = $temp % 64;
$hour = int ($temp / 64);

return ($hour, $min, $sec);
}

# ********************************************************
# help_text
# ********************************************************

# Display the help file

sub help_text
{
print<< "EOT";
Program usage: dirparser [-i input_file]
               dirparser {[-h] | [-v] | [-t]}

 where: -i is the input file name
        -h prints this help file
        -v displays the program version number
        -t prints the directory template file

If no -i switch is provided, the program will ask for the
input file name.

The input file should contain the contents of a FAT directory.
The file should be a binary file that is between 1 and 64 sectors
in length, and is an integral number of sectors. The best way to
obtain a FAT directory is to export the file from your forensics
software.

EOT
return;
}

# ********************************************************
# is_the_rest_blank
# ********************************************************

# This routine checks the remainder of the string to determine
# if it is blank-filled (return 1) or not (return 0) 

sub is_the_rest_blank
{
my ($start,$len,@array) = @_;
my ($a, $return_value);

$return_value = 1;

for ($a = $start; $a < $start+$len; $a++)
  {
  if ($array [$a] != 0x20)
    {
    $return_value = 0;
    last;
    }
  }

return ($return_value);
}

# ********************************************************
# list_hex_digits
# ********************************************************

# Given an array name, starting position, and number of digits, print
# out a string of hex digits in the format 0xAA-BB-CC...

sub list_hex_digits
{
my ($start,$n,@array) = @_;
my $i;

for ($i = $start; $i < $start+$n; $i++)
  {
  if ($i == $start)
      { print "0x"; }
    else
      { print "-"; }
  printf ("%02X", $array[$i]);
  }

return;
}

# ********************************************************
# little_endian
# ********************************************************

# Given an array name, starting position, and number of digits,
# calculate the decimal value using little endian (LSB first)

sub little_endian
{
my ($start,$n,@array) = @_;
my ($i, $sum);

$sum = $array[$start];
for ($i=1; $i<$n; $i++)
  { $sum += $array[$start+$i] * (2**($i*8)); }

return $sum;
}

# ********************************************************
# parsecommandline
# ********************************************************

# Parse command line for file name, if present. Query
# user for any missing information

# Return $state = 1 to indicate that the program should stop
# immediately (switch -h)

sub parsecommandline
{
my $state = 0;

# Parse command line switches ($ARGV array of length $#ARGV)

if ($#ARGV >= 0)
  { 
  for ($i = 0; $i <= $#ARGV; $i++)
    {
    PARMS:
      {
      $ARGV[$i] eq "-i" && do
         {
         $source = $ARGV[$i+1];
         $i++;
         last PARMS;
         };
      $ARGV[$i] eq "-t" && do
         {
         $layout = "layout_dir.txt";
         open (INFILE, "<", $layout)
          or die "Can't open input file \"$layout\": $!";
         while (not (eof INFILE))
           {
           $line = <INFILE>;
           print $line;
           }
         close INFILE;
         $state = 1;
         return $state;
         };
      $ARGV[$i] eq "-v" && do
         {
         $state = 1;
         return $state;
         };
      $ARGV[$i] eq "-h" && do
         {
         help_text();
         $state = 1;
         return $state;
         };

      do
         {
         print "Invalid parameter \"$ARGV[$i]\".\n\n";
         print "Usage: dirparser [-i input_file]\n";
         print "       dirparser {[-h] | [-v] | [-t]}\n\n";
         $state = 1;
         return $state;
         };
      };
    };
  };

# Prompt for file names, if not supplied

if ($source eq "")
  {
  print "Enter the input file name: ";
  chomp ($source = <STDIN>);
  print "\n";
  };

return $state;
}

# ********************************************************
# process_LFN
# ********************************************************

# Parse a 32-byte long file name directory entry

sub process_LFN
{
my ($a, $b, $lfn_len, $new_entry, $file_name, $fn_len, $lpos, $lbase, $temp, $check, $chk_ok);

# A lot of the work here is with global variables!!!

# bbase is the current location in dir_file

# The first thing to do is determine the number of LFN entries comprising
# this filename. If byte 0 != 0xE5, the number is just the lower 6 bits of
# byte 0; otherwise, we need to look ahead and count.

if ($dir_file [$bbase] != 0xe5)
    { $lfn_len = $dir_file [$bbase] % 64; }
  else
    {
    $lfn_len = 1;
    while ($dir_file [$bbase + $lfn_len*32 + 11] == 0x0f)
      { $lfn_len++; }
    }

$new_entry = $entry + $lfn_len - 1;
if ($lfn_len == 1)
    { printf ("*** Entry %d (bytes %d-%d) contains a Long File Name ***\n", $entry, $bbase, $bbase+($lfn_len*32)-1); }
  else
    { printf ("*** Entries %d-%d (bytes %d-%d) contain a Long File Name ***\n", $entry, $new_entry, $bbase, $bbase+($lfn_len*32)-1); }

# Compute checksum. The checksum appears in byte 13 of each LFN entry and is computed from the short file name, which starts
# $lfn_len entries [i.e., ($lfn_len x 32) bytes] after the first LFN entry (byte).

$a = $bbase + ($lfn_len * 32);

# Initial checksum value is the first character of SFN

$check = $dir_file [$a];

# If the SFN is deleted, then the firt character has to be the first valid character in the name, which we need to get from
# the LFN *and* convert to upper case if it is in lower case....

if ($check == 0xE5)
  {
  $check = $dir_file [$bbase + ($lfn_len-1)*32 +1];
  if ($check >= 0x61 && $check <= 0x7A)
    { $check = $check - 0x20; }
  }

# For the next ten rounds, rotate $check one bit and add next characters in SFN

for ($b = 1; $b < 11; $b++)
  {
  $temp = $check % 2;
  $check = ( int($check/2) + ($temp*128) + $dir_file [$a+$b] ) % 256;
  }

$chk_ok = 0; $fn_len = 0;
for ($a = 0; $a < $lfn_len; $a++)
  {
# Use loop to print Entry # X (LFN part Y)...

  printf ("Entry %d (LFN name part %d):\n", $entry+$a, $lfn_len-$a);

  $lbase = $bbase + $a * 32;
  for ($b = 0; $b < 2; $b++)
    { set_output ($lbase+$b*16,16,@dir_file); } 

# transfer UTF-16 characters from dir_file to $file_name
# stop processing the first LFN entry (i.e., when $a==0) either when we get to the 13th character
# or when the next character is 0x00-00 (end-of-name flag)

# $file_name will have ($lfn_len-1)*13 characters plus how many ever characters are in the "first"
# part of the name (i.e., if LFN Part 1 has 3 characters and there are 4 LFN parts, then the
# file name will be 4*13+3 (55) characters in length

  $lpos = ($lfn_len - 1 - $a) * 13;
  for ($b = 0; $b < 13; $b++)
    {
    $temp = $lbase + $lfn_pos [$b];
    if ($dir_file [$temp] == 0x00 && $dir_file [$temp+1] == 0x00)
        {
        last; # Stop processing when the end-of-name indicator (0x00-00) is detected
        }
      else
        {
        $file_name [$lpos+$b] = chr (little_endian ($temp,2,@dir_file));
        $fn_len++;
        }
    }

# Check if the checksum value matches the calculated value...

  if ($dir_file [$lbase+13] != $check)
    { $chk_ok = 1; }
  }

print "      File name: ";
for ($a = 0; $a < $fn_len; $a++)
  { print $file_name [$a]; }

if ($dir_file[$bbase] == 0xe5)
  { print "     (Deleted)"; }
print "\n";

print "      Attributes (0x0F): Long file name\n";
printf ("      Checksum (0x%02X): ", $check);
if ($chk_ok == 0)
    { 
    if ($lfn_len == 1)
        { print "Checksum in LFN entry matches calculated checksum"; }
      else
        { print "Checksum in all LFN entries match calculated checksum"; }
    }
  else
    { print "Checksum mismatch in LFN entries and/or calculated checksum"; }
print "\n\n";

return ($new_entry);
}

# ********************************************************
# process_SFN
# ********************************************************

# Parse a 32-byte short file name directory entry

sub process_SFN
{
my ($dir_entry) = @_;
my ($a, $b, $flag, $temp, $mon, $day, $year, $hour, $min, $sec, $date, $time);

# Byte 0 can be either 0xE5 to indicate a deleted entry (replace
# character with a "_") or an ASCII character.

printf ("*** Entry %d (bytes %d-%d) contains a Short File Name ***\n", $entry, $bbase, $bbase+31);
for ($a=0; $a<2; $a++)
  { set_output ($bbase+$a*16,16,@dir_file); } 

# Together, bytes 0-7 are the file name and 8-10 are the
# extension; the "." is implied.

print "   File name: ";

if ($dir_file [$bbase] == 0xE5)
    { print "?"; }
  else
    { print chr ($dir_file [$bbase]); }

# The file name continues in bytes 1-7, followed by the extension
# in bytes 8-10. If a space (0x20) is found in the name, check to
# ensure that the rest of the name or extension is also filled
# with spaces; it is possible to have an SFN with embedded
# spaces although this violates DOS convention...

# If the entry is for a directory (attribute bit 5 set) or a volume
# label (attribute bit 4), do not print the dot or look at the extension

$temp =0;
if (process_SFN_name ($bbase+1,7,@dir_file) == 1)
  { $temp = 1; }
$dir_flag = (int ($dir_file [$bbase+11] / 16)) %2;
$vol_flag = (int ($dir_file [$bbase+11] / 8)) %2;

if ($dir_flag == 0 && $vol_flag == 00)
  {
  print ".";
  if (process_SFN_name ($bbase+8,3,@dir_file) == 1)
    { $temp = 1; }
  }

if ($dir_file [$bbase] == 0xE5)
  { print "     (Deleted)"; }
if ($temp == 1)
  { print " [Embedded spaces in filename violates SFN convention]"; }
print "\n";

# Byte 11 contains the attribute flags

$temp = $dir_file [$bbase+11];
printf ("   Attributes (0x%02X): ", $temp);
if ($temp == 0x00)
    { print "(None)\n"; }
  else
    {
    $a = 0;
    for ($b=0; $b<8; $b++)
      {
      if ($temp%2 == 1)
        {
        if ($a > 0)
          { print ", "; }
        $flag = sprintf ("0x%02X", 2**$b);
        if (defined $attribute_values{$flag})
            { print "$attribute_values{$flag}"; }
          else
            { print "(Not listed)"; }
        $a++;
        }
      $temp = int ($temp / 2);
      }
    print "\n";
    }

# Byte 12 is reserved and should be set to 0

$temp = $dir_file [$bbase+12];
printf ("   Reserved byte (0x%02X): %d", $temp, $temp);
if ($temp != 0)
  { print " (This byte is reserved and should be set to 0)"; }
print "\n";

# Byte 13 is the C-time in tenths of seconds
# Bytes 14-15 are the C-time, 16-17 the C-date

$time = little_endian ($bbase+14,2,@dir_file);
($hour, $min, $sec) = get_time ($time);
$date = little_endian ($bbase+16,2,@dir_file);
($mon, $day, $year) = get_date ($date);

print "       Creation date (";
list_hex_digits ($bbase+13, 5, @dir_file);
printf("): %02d/%02d/%04d  %02d:%02d:%02d.%03d (Unused in DOS)\n",
  $mon,$day,$year,$hour,$min,$sec, $dir_file [$bbase+13]);

# Bytes 18-19 the A-date

$date = little_endian ($bbase+18,2,@dir_file);
($mon, $day, $year) = get_date ($date);

print "         Access date (";
list_hex_digits ($bbase+18, 2, @dir_file);
printf("         ): %02d/%02d/%04d               (Unused in DOS)\n",
   $mon,$day,$year);

# Bytes 22-23 are the M-time, 24-25 the M-date

$time = little_endian ($bbase+22,2,@dir_file);
($hour, $min, $sec) = get_time ($time);
$date = little_endian ($bbase+24,2,@dir_file);
($mon, $day, $year) = get_date ($date);

printf "   Modification date (";
list_hex_digits ($bbase+22, 4, @dir_file);
printf("   ): %02d/%02d/%04d  %02d:%02d:%02d\n",
  $mon,$day,$year,$hour,$min,$sec);

# Bytes 20-21 are the high-order 16 bits of the starting cluster and
# bytes 26-27 are the low-order 16 bits

$temp = little_endian ($bbase+20,2,@dir_file) * 2**16 +
   little_endian ($bbase+26,2,@dir_file);

print "   Starting cluster (";
list_hex_digits ($bbase+20, 2, @dir_file);
print " ";
list_hex_digits ($bbase+26, 2, @dir_file);
printf("): %s\n", commify ($temp));

# Bytes 28-31 contain the length

print "   File length (";
list_hex_digits ($bbase+28, 4, @dir_file);
printf("): %s bytes\n\n", commify (little_endian ($bbase+28,4,@dir_file)));

return;
}

# ********************************************************
# process_SFN_name
# ********************************************************

# This mess :-) is here just to check on the file name and
# extension to ensure that there is not an embedded space in
# the SFN, which would be a DOS violation.

# return 1 to indicate a DOS error condition; else 0

sub process_SFN_name
{
my ($start,$len,@dir_entry) = @_;
my ($a, $return_value);

$return_value = 0;

for ($a = $start; $a < $start+$len; $a++)
  {
  if ($dir_entry[$a] != 0x20)
      { printf ("%s", chr($dir_entry[$a])); }
    else
      {
      if (is_the_rest_blank ($a,$len-$a+1,@dir_entry) == 1)
          { last; }
        else
          {
          print " ";
          $return_value = 1;
          }
      }
  }

return ($return_value);
}

# ********************************************************
# set_output
# ********************************************************

# Print output of the data bytes to show both hex and ASCII
# values.

sub set_output
{
my ($bpos,$n,@array) = @_;
my ($j, $k, $linemax);

# Format for a maximum of 16 items per line

$linemax = 16;

printf ("%05d: ", $bpos);
for ($j=0; $j<$n; $j++)
  {
  printf (" %02X", $array[$bpos+$j]);
  }
if ($n < $linemax)
  {
  for ($j=1; $j<=$linemax-$n; $j++)
    { print "   "; };
  }

print "   ";
for ($j=0; $j<$n; $j++)
  {
  $k = $array[$bpos+$j];

#   Print a dot (.) for unprintable characters

  if ($k <= 0x1f || ($k >= 0x7f && $k <= 0x9f))
      { print "."; }
    else
      { print chr($k); }
  }
print "\n";

return;
}

